/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        listfctn.c
 * version:     0.1
 * date:        july, 22. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
      
#include "qmc.h"
#include "listfctn.h"


#define ERR_CODE_BASE	200

/*----------------------------------------------------
 * Returns the Idx of the signal named Identifier
 * If no signal of this name exists, -1 is returned
 *---------------------------------------------------------*/
int GetSignalIdx(char *Identifier)
{
  register SIGNAL_LIST_ELEMENT *ListElement;
  
  if((ListElement = StartOfListOfSignals) == NULL) return(-1);
  while(ListElement != NULL) {
    if(strcmp(Identifier, ListElement->Identifier) == 0) return(ListElement->Idx);
    ListElement = ListElement->Successor;
  }
  return(-1);
}

/*----------------------------------------------------
 * Returns the name of the signal with the given index
 * If no signal with this index exists, NULL is returned
 *---------------------------------------------------------*/
char *GetSignalName(int Idx) {
  register SIGNAL_LIST_ELEMENT *ListElement;
  
  if((ListElement = StartOfListOfSignals) == NULL) return(NULL);
  while(ListElement != NULL) {
    if(Idx == ListElement->Idx) return(ListElement->Identifier);
    ListElement = ListElement->Successor;
  }
  return(NULL);
}


/*-----------------------------------------------------
 * Returns the dimension (number of don't cares) of the given 
 * don't care variable
 *-------------------------------------------------------------*/
int GetDontCareDimension(unsigned long DontCare) {
  register int i;
  register int Dim;
  register unsigned long Mask;
  
  if(DontCare == 0) return(0);
  Dim = 0;
  Mask = 1l;
  for(i = 0; i < NumberOfSignals; i++) {
    if((Mask & DontCare) != 0) Dim++;
    Mask <<= 1;
  }
  return(Dim);
}

/*----------------------------------------------------------
 * Checks if the given min term is included or includes previous
 * min terms.
 * Returns line number of the other min term. Zero is returned if no
 * term is found.
 *--------------------------------------------------------------------*/
int CheckMinTerm(MIN_TERM_LIST_ELEMENT *TermPtr) {

  int Idx;
  MIN_TERM_LIST_ELEMENT *ListElement;
  
  for(Idx = 0; Idx < MAX_NUMBER_OF_SIGNALS; Idx++) {
    ListElement = MinTerms[Idx].StartOfListOfMinTerms;
    while(ListElement != NULL) {
      if(CheckTerms(TermPtr->DontCares,     TermPtr->MinTerm,
      		    ListElement->DontCares,ListElement->MinTerm) != 0) {
        return(ListElement->InputLineNo);
      }
      ListElement = ListElement->Successor;
    } /* while */
  } /* for */
  return(0l);
}

/*-----------------------------------------------------
 *
 *----------------------------------------------------------*/
int CheckTerms(unsigned long DontCare1, unsigned long MinTerm1,
	       unsigned long DontCare2, unsigned long MinTerm2 ) {
   if((MinTerm1 | DontCare2) == (MinTerm2 | DontCare1)) return(1);
   return(0);
} 

/*-----------------------------------------------------
 *
 *----------------------------------------------------------*/
static void CopyToArray(char *Array, int Idx) {
  register int ArrayIdx, i;
  register char Mask;

  for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
    Mask = 1;
    for(i = 0; i < 8; i++) {
      if(((8 * ArrayIdx) + i) == Idx) { 
        *Array |= Mask;
        return;
      }
      Mask <<= 1;          
    } /* for(i.. */
    Array++;
  } /* for(ArrayIdx... */
}

/*----------------------------------------------------
 * Allocate IdxArray for all min terms which are no don't cares.
 * Insert the contents of ->Flag (which contains the initial term index. 
 * Set ->Flag to zero.
 *-----------------------------------------------------------------*/
void StartIdx2IdxArray(void) {

  register int Idx; 
  char *NewArray;
  MIN_TERM_LIST_ELEMENT *ListElement;
      
  for(Idx = 0; Idx < MAX_NUMBER_OF_SIGNALS; Idx++) {
    ListElement = MinTerms[Idx].StartOfListOfMinTerms;                    
    while(ListElement != NULL) {
      if(ListElement->Flag != DON_T_CARE_FLAG) {   /* no don't care */
        /*--------- get new array, elements allready set to zero ----------------*/
        if((NewArray = (char *)calloc((unsigned int)SizeOfIdxArray, sizeof(char))) == NULL) {
            fprintf(stderr,"calloc error in StartIdx2IdxArray()\n");
            exit(ERR_CODE_BASE + 11);
        }
        CopyToArray(NewArray, ListElement->Flag);
        ListElement->IdxArray = NewArray;
        ListElement->Flag = 0;
      }	/* if(ListElement->Flag != DON_T_CARE_FLAG) */
      ListElement = ListElement->Successor;
    } /* while */    
  } /* for */
}
