/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        minimize.c
 * version:     0.1
 * date:        july, 23. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qmc.h"
#include "parserfctn.h"
#include "listfctn.h"

#include "minimize.h"

#ifdef	QMC_DEBUG
#include "debugfctn.h"
#endif

#define ERR_CODE_BASE	300

/*----------------------------------------------------
 *       
 *---------------------------------------------------------------*/
static char *CopyIdxArrays(char *Array1, char *Array2) {
  
  char *NewArray;
  register char *ArrayPtr;
  register int Idx, i;
  register char Mask;

  if((Array1 == NULL) && (Array2 == NULL)) return(NULL);
  if((ArrayPtr = NewArray = (char *)calloc((unsigned int)SizeOfIdxArray, sizeof(char))) == NULL) {
    fprintf(stderr,"calloc error (1) in CopyIdxArrays()\n");
    exit(ERR_CODE_BASE + 11);
  }
  if(Array2 == NULL) {
    for(Idx = 0; Idx < SizeOfIdxArray; Idx++) {
      Mask = 1;
      for(i = 0; i < 8; i++) {
        if((*Array1 & Mask) != 0) *ArrayPtr |= Mask;
        Mask <<= 1;
      } /* for(i...*/
      ArrayPtr++;
      Array1++;
    } /* for(Idx... */
    return(NewArray);
  }
  if(Array1 == NULL) {
    for(Idx = 0; Idx < SizeOfIdxArray; Idx++) {
      Mask = 1;
      for(i = 0; i < 8; i++) {
        if((*Array2 & Mask) != 0) *ArrayPtr |= Mask;
        Mask <<= 1;
      } /* for(i...*/
      ArrayPtr++;
      Array2++;
    } /* for(Idx... */
    return(NewArray);
  }
  for(Idx = 0; Idx < SizeOfIdxArray; Idx++) {
    Mask = 1;
    for(i = 0; i < 8; i++) {
      if(((*Array1 & Mask) != 0) || ((*Array2 & Mask) != 0)) *ArrayPtr |= Mask;
      Mask <<= 1;
    } /* for(i...*/
    ArrayPtr++;
    Array1++; Array2++;
  } /* for(Idx... */
  return(NewArray);   
}  

/*----------------------------------------------------
 * 
 *-----------------------------------------------------------------*/
static void InsertMinTerm(MIN_TERM_LIST_ELEMENT *MinTermPtr) {
  int Idx;
  
  if(MinTermPtr == NULL) return;
  if((Idx = GetDontCareDimension(MinTermPtr->DontCares)) > 
 						(MAX_NUMBER_OF_SIGNALS - 1))  {
    fprintf(stderr,"internal error in InsertMinTerm()\n");
    exit(ERR_CODE_BASE + 21);
  }
  if((MinTerms[Idx].StartOfListOfMinTerms == NULL) || (MinTerms[Idx].EndOfListOfMinTerms == NULL)) {
    MinTermPtr->Successor = MinTermPtr->Predecessor = NULL;
    MinTerms[Idx].EndOfListOfMinTerms = MinTerms[Idx].StartOfListOfMinTerms = MinTermPtr;
  } else {
    MinTermPtr->Successor			 = NULL;
    MinTermPtr->Predecessor		   	 = MinTerms[Idx].EndOfListOfMinTerms;
    MinTerms[Idx].EndOfListOfMinTerms->Successor = MinTermPtr;
    MinTerms[Idx].EndOfListOfMinTerms	 	 = MinTermPtr;
  }
}

#if 0
/*---------------------------------------------------------
 * 
 *---------------------------------------------------------------*/
unsigned long DifferenceTerm(unsigned long Term1, unsigned long Term2) {

  int i;
  unsigned long Mask;
  unsigned long DiffTerm;
      
  Mask = 1l;
  DiffTerm = 0l;
  for(i = 0; i < NumberOfSignals; i++) {
    if((Mask & Term1) != (Mask & Term2)) DiffTerm |= Mask;
    Mask <<= 1;
  }
  return(DiffTerm);
}


/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
static int IsDifferenceOfOneBit(unsigned long Term1, unsigned long Term2) {
  int i;
  unsigned long Mask;
  int Difference;
  
  Mask = 1l;
  Difference = 0;
  for(i = 0; i < NumberOfSignals; i++) {
    if((Mask & Term1) != (Mask & Term2)) Difference++;
    Mask <<= 1;
  }
  if(Difference == 1) return(1);
  return(0);
}
#endif



/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
static void Combine(MIN_TERM_LIST_ELEMENT *Term1, MIN_TERM_LIST_ELEMENT *Term2) {
  MIN_TERM_LIST_ELEMENT *NewTerm;
  unsigned long DiffTerm;

  if((NewTerm = (MIN_TERM_LIST_ELEMENT *)malloc(sizeof(MIN_TERM_LIST_ELEMENT))) == NULL) {
    fprintf(stderr,"malloc error Combine()\n");
    exit(ERR_CODE_BASE + 31);
  }
  DiffTerm = (Term1->MinTerm ^ Term2->MinTerm);
  NewTerm->MinTerm   	= ((Term1->MinTerm)   | DiffTerm);
  NewTerm->DontCares 	= ((Term1->DontCares) | DiffTerm);
  NewTerm->InputLineNo 	= -1;
  NewTerm->Value	= 0;
  if((NewTerm->IdxArray	= CopyIdxArrays(Term1->IdxArray, Term2->IdxArray)) == NULL) {
    NewTerm->Flag	= DON_T_CARE_FLAG;
  } else {
    NewTerm->Flag	= 0;
  }
  InsertMinTerm(NewTerm);
}



/*----------------------------------------------------
 * Return non zero value, if the two min terms can be combined to one term.
 * 
 *---------------------------------------------------------------*/
static int IsCombinable(MIN_TERM_LIST_ELEMENT *Term1, MIN_TERM_LIST_ELEMENT *Term2) {
register unsigned long DiffTerm;

  if(Term1->DontCares != Term2->DontCares) return(0);
  if(Term1->MinTerm   == Term2->MinTerm)   return(0);
  
  DiffTerm = (Term1->MinTerm ^ Term2->MinTerm);
  while(DiffTerm != 0) {
    if((DiffTerm & 1l) == 1) {
      if((DiffTerm >> 1) == 0) return(1); 
      else return(0);
    }
    DiffTerm >>= 1;
  } /* while */
  return(0);
}

/*---------------------------------------------------------
 *
 *---------------------------------------------------------------*/
static int AreMinTermesCombinable(unsigned long Term1, unsigned long Term2) {
  register unsigned long DiffTerm;
  
  if(Term1 == Term2) return(1);
  DiffTerm = (Term1 ^ Term2);
  while(DiffTerm != 0) {
    if((DiffTerm & 1l) == 1) {
      if((DiffTerm >> 1) == 0) 	return(1); 
      else 			return(0);
    }
    DiffTerm >>= 1;
  } /* while */
  return(1);
}

/*-------------------------------------------------------
 * Marks all DontCares, whose min terms are not equal respectivly do differ
 * in more than one bit from the min terms of the other terms (no dont cares)
 *--------------------------------------------------------------------------*/
static void MarkUnusedDontCares(void) {
  int OuterDim, InnerDim;
  MIN_TERM_LIST_ELEMENT *OuterListPtr, *InnerListPtr;
  register int MarkedTerms;
  
  MarkedTerms = 0;
  /*------------ mark all don't cares to be cleard ----------------------*/
  for(OuterDim = 0; OuterDim < NumberOfSignals; OuterDim++) {  
    OuterListPtr = MinTerms[OuterDim].StartOfListOfMinTerms;
    while(OuterListPtr != NULL) {
      if(OuterListPtr->Flag == DON_T_CARE_FLAG) OuterListPtr->Flag = CLR_FLAG;
      MarkedTerms++;
      OuterListPtr = OuterListPtr->Successor;
    } /* while(OuterListPtr != NULL) */
  } /* for(OuterDim ... */  
  /*------------- check which don't cares should not be cleared ----------*/
  for(OuterDim = 0; OuterDim < NumberOfSignals; OuterDim++) {
    OuterListPtr = MinTerms[OuterDim].StartOfListOfMinTerms;
    while(OuterListPtr != NULL) {
      if(OuterListPtr->Flag == CLR_FLAG) {	
        for(InnerDim = OuterDim; InnerDim < NumberOfSignals; InnerDim++) {
          InnerListPtr = MinTerms[InnerDim].StartOfListOfMinTerms;
          while(InnerListPtr != NULL) {
            if(InnerListPtr->Flag != CLR_FLAG) {
              if(AreMinTermesCombinable(OuterListPtr->MinTerm,InnerListPtr->MinTerm)) {
                OuterListPtr->Flag = DON_T_CARE_FLAG;
                MarkedTerms--;
              } /* if(OuterListPtr->MinTerm != InnerListPtr->MinTerm) */
            } /* if(InnerListPtr->Flag != DON_T_CARE_FLAG) */
            InnerListPtr = InnerListPtr->Successor;
          } /* while(InnerListPtr != NULL) */
        } /* for(InnerDim ... */
      } /* if(OuterListPtr->Flag == CLR_FLAG) */
      OuterListPtr = OuterListPtr->Successor;      
    } /* while(OuterListPtr != NULL) */
  } /* for(Dimension...*/
}


/*------------------------------------------------------
 *
 *---------------------------------------------------------------*/
static int EqualIdxArrays(char *Array1, char *Array2) {
  register int ArrayIdx;

  if((Array1 == NULL) && (Array2 == NULL)) return(1);
  if((Array1 == NULL) || (Array2 == NULL)) return(0);
  for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
    if(*Array1++ != *Array2++) return(0);
  }
  return(1);
}


/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
static void MarkDontCares(int Dimension) {
  MIN_TERM_LIST_ELEMENT *List;

  List = MinTerms[Dimension].StartOfListOfMinTerms;
  while(List != NULL) {
    if((List->Flag == DON_T_CARE_FLAG) ||
    	(List->IdxArray == NULL)	  ) List->Flag = CLR_FLAG;
    List = List->Successor;
  } /* while(List != NULL) */
}        

/*----------------------------------------------------
 * 
 *---------------------------------------------------------------*/
static void AddIdxArray(char *To, char *From) {
  register int ArrayIdx;
    
  if((To == NULL) || (From == NULL)) return;
  for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
    *To++ |= *From++;
  } /* for(ArrayIdx... */
}


/*----------------------------------------------------
 * Mark min terms with equal DontCares, MinTerm and IdxArray
 *---------------------------------------------------------------*/
static void MarkDuplicates(int Dimension) {
  MIN_TERM_LIST_ELEMENT *OuterList, *InnerList;
    
  OuterList = MinTerms[Dimension].StartOfListOfMinTerms;
  while(OuterList != NULL) {
    InnerList = OuterList->Successor;
    while(InnerList != NULL) {
      if((OuterList->DontCares == InnerList->DontCares) &&
         (OuterList->MinTerm   == InnerList->MinTerm)   	)	{
        if(EqualIdxArrays(OuterList->IdxArray, InnerList->IdxArray) == 0) {
 	  AddIdxArray(OuterList->IdxArray, InnerList->IdxArray);
 	} /* if(EqualIdxArrays() ==  0) */
 	OuterList->Flag = CLR_FLAG;
      } /* if((OuterList->DontCares == ... */
      InnerList = InnerList->Successor;
    } /* while(InnerList != NULL) */  
    OuterList = OuterList->Successor;
  } /* while(OuterList != NULL) */
}


/*-----------------------------------------------------
 *
 *----------------------------------------------------------*/
static int IsInclosed(char *ArrayNplus1, char *ArrayN) {
  register int ArrayIdx, i;
  register char Mask;

  if(ArrayN 	 == NULL) return(1);
  if(ArrayNplus1 == NULL) return(0);   
  for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
    Mask = 1;
    for(i = 0; i < 8; i++) {
      if((*ArrayN & Mask) != 0) {
        if((*ArrayNplus1 & Mask) == 0) return(0);
      }
      Mask <<= 1;          
    } /* for(i.. */
    ArrayNplus1++;
    ArrayN++;        
  } /* for(ArrayIdx... */
  return(1);
}

/*----------------------------------------------------
 * Mark min terms whose DontCares, MinTerm and IdxArray variables
 * are a subset of other min terms in the lower dimensions.
 *---------------------------------------------------------------*/
static void MarkEnclosedTermes(int Dimension) {
  MIN_TERM_LIST_ELEMENT *DimensionN, *DimensionNplus1;
  
  int LowerDim;
  
  DimensionN = MinTerms[Dimension].StartOfListOfMinTerms;
  while(DimensionN != NULL) {
    for(LowerDim = Dimension + 1; LowerDim < MAX_NUMBER_OF_SIGNALS; LowerDim++) {
      DimensionNplus1 = MinTerms[LowerDim].StartOfListOfMinTerms;
      while(DimensionNplus1 != NULL) {
        if(IsInclosed(DimensionNplus1->IdxArray, DimensionN->IdxArray)) {
          DimensionN->Flag = CLR_FLAG;
        }
        DimensionNplus1 = DimensionNplus1->Successor;
      } /* while(DimensionNplus1 != NULL) */
    } /* for(LowerDim ... */
    DimensionN = DimensionN->Successor;
  } /* while(DimensionN != NULL) */
  return;
}


/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
void ClearMarkedElements(void) {
  MIN_TERM_LIST_ELEMENT *List;
  MIN_TERM_LIST_ELEMENT *DeleteElement, *Successor, *Predecessor;
  register int Dim;
  
  for(Dim = 0; Dim < NumberOfSignals; Dim++) {
    /*-------- clear all marked elements at the begin of the list ----------------*/
    while(MinTerms[Dim].StartOfListOfMinTerms != NULL) {
      if((MinTerms[Dim].StartOfListOfMinTerms)->Flag != CLR_FLAG) break;
      DeleteElement = MinTerms[Dim].StartOfListOfMinTerms;
      if(((MinTerms[Dim].StartOfListOfMinTerms) = DeleteElement->Successor) != NULL) {
        (MinTerms[Dim].StartOfListOfMinTerms)->Predecessor = NULL;
      }
      if(DeleteElement->IdxArray != NULL) free(DeleteElement->IdxArray);
      free(DeleteElement); 
    } /* while((MinTerms... */
    if(MinTerms[Dim].StartOfListOfMinTerms == NULL) {
      MinTerms[Dim].EndOfListOfMinTerms = NULL;
    }
    /*------------ clear marked elements from the middle of the list --------------*/
    List = MinTerms[Dim].StartOfListOfMinTerms;
    while(List != NULL) {
      if(List->Flag == CLR_FLAG) {
        DeleteElement = List;
        Successor     = List->Successor;
        Predecessor   = List->Predecessor;
        List = Successor;
        if(Predecessor == NULL) { 
          MinTerms[Dim].StartOfListOfMinTerms = Successor;
        } else {  
          Predecessor->Successor = Successor;
        }
        if(Successor == NULL) { 
          MinTerms[Dim].EndOfListOfMinTerms = Predecessor;
        } else {
          Successor->Predecessor = Predecessor;
        }
        if(DeleteElement->IdxArray != NULL) free(DeleteElement->IdxArray);
        free(DeleteElement);
      } else {
        List = List->Successor;
      }
    } /* while(List != NULL) */
    if(MinTerms[Dim].StartOfListOfMinTerms == NULL) MinTerms[Dim].EndOfListOfMinTerms = NULL;
  } /* for(Dim... */
}


/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
void Minimize(void) {
  int Idx;
  MIN_TERM_LIST_ELEMENT *OuterListPtr, *InnerListPtr;

#if 0
  MarkUnusedDontCares(); 
#endif  
  for(Idx = 0; Idx < (NumberOfSignals - 1); Idx++) {
    if(Silent == 0) { 
      printf("%d. of %d minimization steps\r",Idx +1, NumberOfSignals);
      fflush(stdout);
    }
    MarkDuplicates(Idx);	
    ClearMarkedElements();
    OuterListPtr = MinTerms[Idx].StartOfListOfMinTerms;
    while(OuterListPtr != NULL) {
      InnerListPtr = OuterListPtr->Successor;
      while(InnerListPtr != NULL) {
        if(IsCombinable(InnerListPtr, OuterListPtr)) {
          Combine(OuterListPtr, InnerListPtr);
        }
        InnerListPtr = InnerListPtr->Successor;
      } /* while(InnerListPtr != NULL) */
      OuterListPtr = OuterListPtr->Successor;
    } /* while(OuterListPtr != NULL) */
    MarkDontCares(Idx);	
    MarkEnclosedTermes(Idx);	
  } /* for */
  MarkDuplicates(Idx);	
  ClearMarkedElements();
  if(Silent == 0) { 
    printf("minimization done...       \n");
    fflush(stdout);
    }
}
