/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        output.c
 * version:     0.1
 * date:        july, 25. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>         /* time() and localtime() */
      
#include "qmc.h"
#include "output.h"

#define ERR_CODE_BASE 	500

/*-----------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static void PrintIdxLine(FILE *Fp, char *Array) {

int Idx, i;
char Mask;

  if(Array == NULL) {
    fprintf(Fp, "<null>");
    return;
  } 
  for(Idx = 0; Idx < SizeOfIdxArray; Idx++) {
    Mask = 1;
    for(i = 0; i < 8; i++) {
      if((Idx > 0) || (i > 0)) {
        if((Array[Idx] & Mask) != 0) fprintf(Fp, "X");
        else fprintf(Fp, "-");
      }
      Mask <<= 1;
      if(((Idx * 8) + i) >= MinTermIdx) return;
    }
  }
}

/*-----------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static char *GetSignalName(int Idx)
{
  SIGNAL_LIST_ELEMENT *ListElement;
  static char ErrMsg1[] = "<list-not-initialized>";
  static char ErrMsg2[] = "<element not found>";
  
  if((ListElement = StartOfListOfSignals) == NULL) {
    return(ErrMsg1);
  }
  while(ListElement != NULL) {
    if(ListElement->Idx == Idx) { 
      return(ListElement->Identifier);
    }
    ListElement = ListElement->Successor;
  }
  return(ErrMsg2);
}


static void FillWithBlank(FILE *Fp, unsigned int i) {
  while(i > 0) {
    fprintf(Fp, " "); 
    i--;
  }
}

/*-----------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static void PrintMinTermDnf(FILE *Fp, unsigned long DontCare, unsigned long MinTerm) {

  int i;
  int FirstSignal;
  unsigned long Mask;
  
  Mask = 1l;
  FirstSignal = 1;
  for(i = 0; i < NumberOfSignals; i++) {
    if((DontCare & Mask) == 0) { 
      if(FirstSignal != 0) {
        FirstSignal = 0;
      } else {
        fprintf(Fp, " %s ",AND_STR);
      }
      if((MinTerm & Mask) == 0) fprintf(Fp, "%s",NEGATION_STR);
      else FillWithBlank(Fp, strlen(NEGATION_STR));
      fprintf(Fp, "%s", GetSignalName(i));
    } else {
      FillWithBlank(Fp, strlen(AND_STR) + 2 + strlen(NEGATION_STR) + strlen(GetSignalName(i)));
    } /* if -- else */
    Mask <<= 1;
  } /* for */
}


/*-----------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static void PrintSignalName(FILE *Fp)
{
  SIGNAL_LIST_ELEMENT *ListElement;
  
  if((ListElement = StartOfListOfSignals) == NULL) {
    return;
  }
  fprintf(Fp, "%s\n", SIGNALS_STR);
  while(ListElement != NULL) {
    fprintf(Fp, "\t%s\n", ListElement->Identifier);
    ListElement = ListElement->Successor;
  }
  return;
}

/*------------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
void WriteOutputHeader(FILE *Fp) {
  time_t		Time;

  fprintf(Fp,"%s file created by '%s' (ver.%s/%s)\n",COMMENT_STR,PrgName, VERSION, DATE);
  fprintf(Fp,"%s Input file: \t%s\n", COMMENT_STR, InputFileName);
  time(&Time);
  fprintf(Fp,"%s created:  \t%s",COMMENT_STR     ,asctime(localtime(&Time)));
  if(NoDontCares != 0) fprintf(Fp,"%s\n%s No input don't cares used!\n",COMMENT_STR, COMMENT_STR);
  fprintf(Fp,"%s\n",COMMENT_STR);
  if(NoSymbolList == 0) {
    PrintSignalName(Fp);
  }
}
  
/*------------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
void WriteOutputEquation(FILE *Fp) {
  int                   Dimension;
  int 			FirstLine;
  MIN_TERM_LIST_ELEMENT *MinTermListElement;  
  
  fprintf(Fp,"\n%s\n", EQUATION_STR);
  if(OutputPolaritiy == NEGATED) fprintf(Fp,"%s", NEGATION_STR);
  else				 FillWithBlank(Fp,strlen(NEGATION_STR));
  fprintf(Fp,"%s %s ",OutputName, AssignmentStr);
  FirstLine = 1;
  for(Dimension = 0; Dimension < NumberOfSignals; Dimension++) {
    MinTermListElement = MinTerms[Dimension].StartOfListOfMinTerms;
    while(MinTermListElement != NULL) {
      if((MinTermListElement->Flag == SELECT_FLAG) || (MinTermListElement->Flag == ESSENTIAL_FLAG)) {
        if(FirstLine != 0) {
          FirstLine = 0;
        } else {
          FillWithBlank(Fp,strlen(OutputName) + strlen(NEGATION_STR) + strlen(AssignmentStr) - strlen(OR_STR));
          fprintf(Fp," %s ",OR_STR);
        }
        PrintMinTermDnf(Fp, MinTermListElement->DontCares, MinTermListElement->MinTerm);
        fprintf(Fp," %s Idx: ", COMMENT_STR);
        PrintIdxLine(Fp, MinTermListElement->IdxArray);
        fprintf(Fp,"\n");
      } /* if((MinTermListElement->Flag == SELECT_FLAG) || ... */
      MinTermListElement = MinTermListElement->Successor;
    } /* while(MinTermListElement != NULL) */
  }    
}


/*------------------------------------------------------------
 *
 *----------------------------------------------------------------------*/
void WriteAllMinTerms(FILE *Fp) {
  int                   Dimension;
  int 			FirstLine;
  MIN_TERM_LIST_ELEMENT *MinTermListElement;
    
  fprintf(Fp,"\n%sAll remaining min terms:\n", COMMENT_STR);
  if(NoDontCares != 0) fprintf(Fp,"%s No input don't cares used!\n",COMMENT_STR);
  fprintf(Fp,"%s ",COMMENT_STR);
  if(OutputPolaritiy == NEGATED) fprintf(Fp,"%s", NEGATION_STR);
  else				 FillWithBlank(Fp,strlen(NEGATION_STR));
  fprintf(Fp,"%s %s ",OutputName,AssignmentStr);
  FirstLine = 1;
  for(Dimension = 0; Dimension < NumberOfSignals; Dimension++) {
    MinTermListElement = MinTerms[Dimension].StartOfListOfMinTerms;
    while(MinTermListElement != NULL) {
      if(FirstLine != 0) {
        FirstLine = 0;
      } else {
        fprintf(Fp,"%s ", COMMENT_STR);
        FillWithBlank(Fp,strlen(NEGATION_STR) + strlen(OutputName) + strlen(AssignmentStr) - strlen(OR_STR));
        fprintf(Fp," %s ",OR_STR);
      }
      PrintMinTermDnf(Fp, MinTermListElement->DontCares, MinTermListElement->MinTerm);
      fprintf(Fp," %s Idx: ",COMMENT_STR);
      PrintIdxLine(Fp, MinTermListElement->IdxArray);
      fprintf(Fp,"\n");
      MinTermListElement = MinTermListElement->Successor;
    } /* while(MinTermListElement != NULL) */
  }    
}


