/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        parserfctn.c
 * version:     0.1
 * date:        july, 22. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qmc.h"
#include "parserfctn.h"
#include "listfctn.h"

#define ERR_CODE_BASE	100

/*------------------------------------
 * print first part of parser error message
 *------------------------------------------------*/
int yyerror(char *ErrMsg)
  {
    fprintf(stderr,"\n\n*** lexline %d: %s ", yylineno, ErrMsg);
    return(0);
  }

/*------------------------------------
 * print second part of parser error message, 
 * abort program
 *--------------------------------------------------*/
int YaccErrMsg(char *ErrMsg)
  {
    fprintf(stderr,"(%s) ***\n\n",ErrMsg);
    return(0);
  }


/*------------------------------------------- 
 * Put identifier into list of signals to keep the relation of
 * signal name and postition number.
 * Count number of signals
 *---------------------------------------------------------------------*/
void AddSignalId(char *Identifier)
{ 
  register SIGNAL_LIST_ELEMENT	*NewElement;

  if(GetSignalIdx(Identifier) < 0) {
    if((NewElement = (SIGNAL_LIST_ELEMENT *)malloc(sizeof(SIGNAL_LIST_ELEMENT))) == NULL) {
      fprintf(stderr,"*** malloc-error in function AddSignalId() ***\n");
      exit(ERR_CODE_BASE + 1);
    };
    NewElement->Idx 		= NumberOfSignals;
    NewElement->Identifier	= Identifier;
    if((StartOfListOfSignals == NULL) || (EndOfListOfSignals == NULL)) {
      NewElement->Successor = NewElement->Predecessor = NULL;
      StartOfListOfSignals = EndOfListOfSignals = NewElement;
    } else {
      NewElement->Successor 		= NULL;
      NewElement->Predecessor 		= EndOfListOfSignals;
      EndOfListOfSignals->Successor	= NewElement;
      EndOfListOfSignals		= NewElement;
    }
    if(++NumberOfSignals > MAX_NUMBER_OF_SIGNALS) {
      fprintf(stderr,"line %d: too many input signals spezified (max. %d allowed)\n",
      		yylineno, MAX_NUMBER_OF_SIGNALS);
      exit(ERR_CODE_BASE + 2);
    }
  } else {  
    fprintf(stderr,"line %d: duplicate definition of signal name '%s' -- ignored\n",
    		yylineno, Identifier);
    free((void *)Identifier);	
  }
}

/*----------------------------------------------------
 * 
 *-----------------------------------------------------------------*/
void InsertActualMinTerm(void) {
  int Idx;
  int TermLine;

  if(ActualMinTerm == NULL) return;
  if((TermLine = CheckMinTerm(ActualMinTerm)) != 0) {
    fprintf(stderr,"warning: min term in line %d mismatchtes with term in line %d\n",
    		ActualMinTerm->InputLineNo, TermLine);
  }
  if((Idx = GetDontCareDimension(ActualMinTerm->DontCares)) >= MAX_NUMBER_OF_SIGNALS)  {
    fprintf(stderr,"internal error in InsertActualMinTerm()\n");
    exit(ERR_CODE_BASE + 11);
  }
  if((MinTerms[Idx].StartOfListOfMinTerms == NULL) || (MinTerms[Idx].EndOfListOfMinTerms == NULL)) {
    ActualMinTerm->Successor = ActualMinTerm->Predecessor = NULL;
    MinTerms[Idx].EndOfListOfMinTerms = MinTerms[Idx].StartOfListOfMinTerms = ActualMinTerm;
  } else {
    ActualMinTerm->Successor			   = NULL;
    ActualMinTerm->Predecessor		   	   = MinTerms[Idx].EndOfListOfMinTerms;
    MinTerms[Idx].EndOfListOfMinTerms->Successor = ActualMinTerm;
    MinTerms[Idx].EndOfListOfMinTerms	 	   = ActualMinTerm;
  }
  ActualMinTerm = NULL;
}

/*-----------------------------------------------
 *
 *------------------------------------------------------------------*/
void NewMinTerm(int LineNo) {
  
  InsertActualMinTerm();
  /*----------- create new min term ----------------------------------------------*/
  MinTermIdx++;
  if((ActualMinTerm = (MIN_TERM_LIST_ELEMENT *)malloc(sizeof(MIN_TERM_LIST_ELEMENT))) == NULL) {
    fprintf(stderr,"malloc error (1) in NewMinterm()\n");
    exit(ERR_CODE_BASE + 21);
  }
  ActualMinTerm->DontCares 	|= (unsigned long)0xFFFFFFFFl;
  ActualMinTerm->MinTerm	|= (unsigned long)0xFFFFFFFFl;
  ActualMinTerm->InputLineNo	= LineNo;
  ActualMinTerm->IdxArray	= NULL;
  ActualMinTerm->Flag		= MinTermIdx;
  ActualMinTerm->Value 		= 0;
}


/*-----------------------------------------------
 *
 *------------------------------------------------------------------*/
void AddToMinTerm(char Polarity, char *Identifier) {

  int SignalIdx;
  unsigned long Mask;
  
  if(ActualMinTerm == NULL) {
    fprintf(stderr,"error in AddToMinTerm(): ActualMinTerm is not initialized\n");
    exit(ERR_CODE_BASE + 31);
  }
  if((SignalIdx = GetSignalIdx(Identifier)) < 0) {
    fprintf(stderr,"error: cannot find signal '%s' in list of signals\n",Identifier);
    exit(ERR_CODE_BASE + 32);
  }
  Mask = ~(1l << SignalIdx);
  ActualMinTerm->DontCares &= Mask;
  if(Polarity == NEGATED) ActualMinTerm->MinTerm &= Mask;
  free(Identifier);
}

/*----------------------------------------------------
 * 
 *-----------------------------------------------------------------*/
void InsertActualDontCare() {

  int Idx;
  int TermLine;

  if(NoDontCares != 0) return;		/* do not use don't caes (cmd line option) */
  if(ActualDontCare == NULL) return;
  if((TermLine = CheckMinTerm(ActualDontCare)) != 0) {
    fprintf(stderr,"warning: don't care term in line %d mismatchtes with term in line %d\n",
    		ActualDontCare->InputLineNo, TermLine);
  }
  if((Idx = GetDontCareDimension(ActualDontCare->DontCares)) >= MAX_NUMBER_OF_SIGNALS) {
    fprintf(stderr,"internal error in InsertActualDontCare()\n");
    exit(ERR_CODE_BASE + 41);
  }
  if((MinTerms[Idx].StartOfListOfMinTerms == NULL) || (MinTerms[Idx].EndOfListOfMinTerms == NULL)) {
    ActualDontCare->Successor = ActualDontCare->Predecessor = NULL;
    MinTerms[Idx].EndOfListOfMinTerms = MinTerms[Idx].StartOfListOfMinTerms = ActualDontCare;
  } else {
    ActualDontCare->Successor			 = NULL;
    ActualDontCare->Predecessor			 = MinTerms[Idx].EndOfListOfMinTerms;
    MinTerms[Idx].EndOfListOfMinTerms->Successor = ActualDontCare;
    MinTerms[Idx].EndOfListOfMinTerms	 	 = ActualDontCare;
  }
  ActualDontCare = NULL;
}


/*-----------------------------------------------                      
 *
 *------------------------------------------------------------------*/
void NewDontCare(int LineNo) {

  DontCareIdx++;
  if(NoDontCares != 0) return;          /* do not use don't caes (cmd line option) */
  InsertActualDontCare();
  /*----------- create new min term ----------------------------------------------*/
  if((ActualDontCare = (MIN_TERM_LIST_ELEMENT *)malloc(sizeof(MIN_TERM_LIST_ELEMENT))) == NULL) {
    fprintf(stderr,"malloc error (1) in NewDontCare()\n");
    exit(ERR_CODE_BASE + 51);
  }
  ActualDontCare->DontCares	|= (unsigned long)0xFFFFFFFFl;
  ActualDontCare->MinTerm	|= (unsigned long)0xFFFFFFFFl;
  ActualDontCare->InputLineNo	= LineNo;
  ActualDontCare->IdxArray	= NULL;
  ActualDontCare->Flag		= DON_T_CARE_FLAG;
  ActualDontCare->Value 	= 0;
}

/*-----------------------------------------------
 *
 *------------------------------------------------------------------*/
void AddToDontCare(char Polarity, char *Identifier) {

  int SignalIdx;
  unsigned long Mask;
  
  if(NoDontCares != 0) return;          /* do not use don't caes (cmd line option) */
  if(ActualDontCare == NULL) {
    fprintf(stderr,"error in AddToDontCare(): ActualDontCare is not initialized\n");
    exit(ERR_CODE_BASE + 61);
  }
  if((SignalIdx = GetSignalIdx(Identifier)) < 0) {
    fprintf(stderr,"error: cannot find signal '%s' in list of signals\n",Identifier);
    exit(ERR_CODE_BASE + 62);
  }
  Mask = ~(1l << SignalIdx);
  ActualDontCare->DontCares &= Mask;
  if(Polarity == NEGATED) ActualDontCare->MinTerm &= Mask;
  free(Identifier);
}


/*-----------------------------------------------
 *
 *------------------------------------------------------------------*/
void GetOutput(char *Identifier, char Polaritiy) {

  OutputPolaritiy = Polaritiy;
  OutputName = Identifier; 
}


/*-----------------------------------------------
 *
 *------------------------------------------------------------------*/
void CheckOutput(char *Identifier, char Polaritiy) {

  if(OutputPolaritiy != Polaritiy) {
    fprintf(stderr,"error in line %d: definition of output polarity mismatches previous definition\n",
    		yylineno);
    exit(ERR_CODE_BASE + 71);    		
  }
  if(strcmp(OutputName, Identifier) != 0) {
    fprintf(stderr,"error in line %d: definition of output mismatches previous definition\n",
  		yylineno);
    exit(ERR_CODE_BASE + 72);
  }
  free(Identifier);
}
