/***********************************************************************
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#undef  extern
#define extern
#include "qmc.h"
#undef extern

void yyparse(void);


#include "parserfctn.h"
#include "listfctn.h"
#include "minimize.h"
#include "select.h"
#include "output.h"

#ifdef QMC_DEBUG
#include "debugfctn.h"
#endif


void Usage(void);

/*-----------------------------------------------
 *
 *--------------------------------------------------------*/
void Usage(void) {
  fprintf(stderr,"\n%s (ver.%s/%s) -- minimize a logical equation (Quine McCluskey)\n",
  		PrgName, VERSION, DATE);
  fprintf(stderr,"\nusage: %s <dnf-file> [<options>] \n",PrgName);
  fprintf(stderr,"\t<dnf-file>: file, containing one logical equation (dnf)\n");
  fprintf(stderr,"\t<options>:\n");
  fprintf(stderr,"\t%c%s\t<name> name of output file (default <input>%s)\n",CMD_ID, CMD_OUTPUT_FILE, EXTENSION);
  fprintf(stderr,"\t%c\toutput to stdout\n",CMD_ID);
  fprintf(stderr,"\t%c%s\tdo not use don't cares give in the input file\n",CMD_ID, CMD_NO_DONT_CARE);
  fprintf(stderr,"\t%c%s\tdo not print symbol list into output file\n",CMD_ID, CMD_NO_SYSM_LIST);
  fprintf(stderr,"\t%c%s\tdo not add list of minterms to output file\n",CMD_ID,CMD_NOT_ADD_MIN_TERMS);
  fprintf(stderr,"\t%c%s\trun silently\n",CMD_ID,CMD_SILENT);
  exit(1);
}

/*-------------------------------------------------
 *
 *------------------------------------------------------*/
void main(int argc, char *argv[]) 
{
  register int i;
  FILE *OutFile;
  char *ArgPtr;
  
  PrgName = rindex(argv[0],(int)('/'));
  if(PrgName == NULL) PrgName = argv[0];
  else PrgName++;
  /*---------------- analyze command line arguments -----------------*/
  NoDontCares = NoSymbolList = NotAddMinTerms = 0;
  OutFile = yyin = NULL;
  if(argc == 1) {
    yyin = stdin;
    InputFileName = "<stdin>";
    OutFile = stdout;
    OutputFileName = "<stdout>";
    Silent = 1;
  } else {
    for(i = 1; i < argc; i++) {
      if(*argv[i] == CMD_ID) {
        ArgPtr = argv[i];
        ArgPtr++;
        if(*ArgPtr == '\0') {
          OutFile = stdout;
          Silent = 1;
        } else {
          if(strcasecmp(ArgPtr,CMD_OUTPUT_FILE) == 0)  {
            i++;
            if(OutFile != NULL) {
              fprintf(stderr,"multiple output file names specified\n");
              Usage();
            }
            if(i > argc) {
              fprintf(stderr,"no output file names specified\n");
              Usage();
            }
            OutputFileName = argv[i];
            if((OutFile = fopen(OutputFileName,"w")) == NULL) {
              fprintf(stderr,"error opening output file '%s'\n", OutputFileName);
              Usage();
            }
          } else {
            if(strcasecmp(ArgPtr, CMD_NO_DONT_CARE) == 0) {
              NoDontCares  = 1;
            } else {	
              if(strcasecmp(ArgPtr, CMD_NO_SYSM_LIST) == 0) {
    	    NoSymbolList = 1;
    	    } else { 
                if(strcasecmp(ArgPtr, CMD_NOT_ADD_MIN_TERMS) == 0) {
                  NotAddMinTerms = 1;
                } else {
                  if(strcasecmp(ArgPtr, CMD_SILENT) == 0) {
                    Silent = 1;
                  } else {
                    fprintf(stderr,"unknown command line argument '%s'\n", argv[i]);
                    Usage();
                  }
                }
              }
            }
          }
        } /* else -- if(*ArgPtr == '\0') */
      } else {
        if(yyin != NULL) {
          fprintf(stderr,"multiple defined input file name\n");
          Usage();
        }
        InputFileName = argv[i];
        if((yyin = fopen(InputFileName,"r")) == NULL) {
          fprintf(stderr,"error opening input file '%s'\n",InputFileName);
          Usage();
        }
      } /* else -- if(*argv[i] == CMD_ID) */
    } /* for(i */
  } /* else -- if(argc == 1) */
  /*---------------- initialize global variables --------------------*/
  if(yyin == NULL) {
    InputFileName = "<stdin>";
    yyin = stdin;
  }
  if(OutFile == NULL) { 
    if((OutputFileName = (char *)calloc(strlen(InputFileName) + strlen(EXTENSION) + 2, sizeof(char))) == NULL) {
      fprintf(stderr,"calloc erron in main()\n");
      exit(2);
    }
    strcpy(OutputFileName, InputFileName);
    strcat(OutputFileName, EXTENSION);
    if((OutFile = fopen(OutputFileName,"w")) == NULL) {
      fprintf(stderr,"error opening output file '%s'\n", OutputFileName);
    }    
  }
  yylineno = 1;
  StartOfListOfSignals 	= EndOfListOfSignals 	= NULL;
  NumberOfSignals	= 0;
  for(i = 0; i < MAX_NUMBER_OF_SIGNALS; i++) {
    MinTerms[i].StartOfListOfMinTerms = NULL;
    MinTerms[i].EndOfListOfMinTerms = NULL;
  }
  DontCareIdx = MinTermIdx = 0;
  SizeOfIdxArray = -1;
  ActualMinTerm = NULL;
  ActualDontCare = NULL;
  AssignmentStr = OutputName = "";
  OutputPolaritiy = 0;
  /*------------------------ message ----------------------------------------*/
  if(Silent == 0) {
    printf("%s (ver.%s/%s) -- minimize file '%s' -> '%s'\n", PrgName, VERSION,
    			DATE, InputFileName, OutputFileName);
    fflush(stdout);
  }
  /*------------------------ parse input file -------------------------------*/
  yyparse();  				/* call the parser 	*/
  if(yyin != stdin) fclose(yyin);
  /*---------------------------------------*/
  SizeOfIdxArray = ((MinTermIdx / 8) + 1);
  StartIdx2IdxArray();		
  if(Silent == 0) {
    printf("%d input terms", MinTermIdx);
    if(NoDontCares == 0) printf(", %d input don't cares\n", DontCareIdx);
    else printf(", no input don't cares used\n");
  }
  /*--------------------- start minimization --------------------------------*/
  Minimize();	
  WriteOutputHeader(OutFile);
  if(NotAddMinTerms == 0) WriteAllMinTerms(OutFile);  
  SelectMinTerms();
  /*------ print contents of internal lists for debugging if desired --------*/
#ifdef QMC_DEBUG 
  printf("Number of Signals: %d\n", NumberOfSignals);
  printf("MinTermIdx: %d, SizeOfIdxArray = %d\n",MinTermIdx, SizeOfIdxArray);
  DebugPrintSignalList();
  DebugPrintMinTermList();
#endif 
  /*----------------------- write output files -------------------------*/
  WriteOutputEquation(OutFile);
  if(OutFile != stdout) fclose(OutFile);
}
