%{
/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        qmc_lex.l
 * version:     0.1
 * date:        july, 22. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 * 
 * description: lexer to read a dnf decription file. Needs flex.
 *
 *********************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "y.tab.h"
#include "qmc.h"

int yywrap(void);
char *IdDup(char *, int);

%}

delimiter	[ \t]
newline		\n
comment_id	[#;]

signals			"SIGNALS"
equation		"EQUATION"
dont_cares		"DON\'T_CARES"
end			"END"
and			"*"
or			"+"
not			"/"
assignment		"="
reg_assign		":="

special_chars	[\-\_\$\@\.]
umlaute		[\\\\\\\]
id_element	[a-zA-Z0-9{special_chars}{umlaute}]
identifier	{id_element}*
asciichar	.
comment_line	{comment_id}{asciichar}*\n

%%

{comment_line}		{yylineno++; }	/* ignore comment lines but count them */
{delimiter}		;		/* ignore whitespace */
{newline}		{yylineno++; }; /* count lines */

{signals}		return(T_SIGNALS);
{equation}		return(T_EQUATION);
{dont_cares}		return(T_DONT_CARES);
{end}			return(T_END);

{and}			return(T_AND); 
{or}			return(T_OR);
{not}			return(T_NOT);
{reg_assign}		return(T_REG_ASSIGNMENT);
{assignment}		return(T_ASSIGNMENT);

{identifier}		{yylval.StringPtr = IdDup(yytext,yyleng); return(T_IDENTIFIER); };

%%

/*---------------------------------------------------------*/
int yywrap(void)
  {
    return(1);
  }

/*-----------------------------------------------------------
 * check for line overflow, duplicate identifier name 
 * in case of error return null pointer.
 *------------------------------------------------------------------------*/
char *IdDup(char *StrPtr, int StrLength)
  {
    StrPtr[StrLength] = '\0'; 	/* force correct termination of the string */
    return(strdup(StrPtr));
  }
