%{
/*********************************************************
 * project:	qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:	qmc_parse.y
 * version:     0.1
 * date:	july, 22. 1995
 * author:	volker bosch (volker.bosch@z.zgs.de)
 *
 * description:	parser to read a dnf decription file. 
 *
 *********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qmc.h" 
#include "parserfctn.h"

#ifdef YYDEBUG
	extern int yydebug = 1;
#endif

/*------------------------------------------------------------------------------*/
%}

%union 	{
	long   		IntVal;
	unsigned short	HexVal;
	int		CharVal;
	void		*StringPtr;
	}

%token 	T_SIGNALS	 	T_EQUATION	T_DONT_CARES
%token	T_AND			T_OR		T_NOT		
%token	T_REG_ASSIGNMENT	T_ASSIGNMENT 
%token	<StringPtr>		T_IDENTIFIER
%token	T_END

%type  <StringPtr>	read_identifier


%start  qmc

%%

qmc			:	define_signals define_equation			 equation_end
			|	define_signals define_equation define_dont_cares dont_care_end
			;

dont_care_end		:	T_END
					{InsertActualDontCare(); }
			;

equation_end		:	T_END
					{InsertActualMinTerm(); }
			;

define_signals		:	T_SIGNALS	signal_list
			;

signal_list		:	read_signal_id
			|	signal_list 	read_signal_id
			;

read_signal_id		:	read_identifier
					{ AddSignalId($1); }
			;

define_equation		:	first_minterm list_of_minterms
			;

first_minterm		:	T_EQUATION       read_direct_equ_output_name  read_equ_assignment minterm
			|	T_EQUATION T_NOT read_negated_equ_output_name read_equ_assignment minterm
			;

read_equ_assignment	:	T_ASSIGNMENT
					{ AssignmentStr = "="; }
			|	T_REG_ASSIGNMENT
					{ AssignmentStr = ":="; }
			;

read_direct_equ_output_name:	read_identifier
					{GetOutput($1,DIRECT); NewMinTerm(yylineno); }
			;

read_negated_equ_output_name:	read_identifier
					{GetOutput($1, NEGATED); NewMinTerm(yylineno); }
			;

minterm			:	                    read_direct_equ_signal
			|	              T_NOT read_negated_equ_signal
			|	minterm	T_AND       read_direct_equ_signal
			|	minterm T_AND T_NOT read_negated_equ_signal
			;

read_direct_equ_signal	:	read_identifier
					{AddToMinTerm(DIRECT, $1); }
			;

read_negated_equ_signal	:	read_identifier
					 {AddToMinTerm(NEGATED, $1); }
			;

list_of_minterms	: 	new_minterm	 minterm
			|	list_of_minterms new_minterm minterm
			;

new_minterm		:	T_OR
					{NewMinTerm(yylineno); }
			;

define_dont_cares	:	first_dont_care list_of_dont_cares
			;

first_dont_care		:	T_DONT_CARES       read_direct_dc_output_name  read_dc_assignment dont_care
					{InsertActualMinTerm(); };
			|	T_DONT_CARES T_NOT read_negated_dc_output_name read_dc_assignment dont_care
					{InsertActualMinTerm(); };
			;

read_dc_assignment	:	T_ASSIGNMENT
			|	T_REG_ASSIGNMENT
			;

read_direct_dc_output_name:	read_identifier
					{CheckOutput($1, DIRECT); NewDontCare(yylineno);}
			;

read_negated_dc_output_name:	read_identifier
					{CheckOutput($1, NEGATED); NewDontCare(yylineno);}
			;

dont_care		:	                      read_direct_dc_signal
			|	                T_NOT read_negated_dc_signal
			|	dont_care T_AND       read_direct_dc_signal
			|	dont_care T_AND T_NOT read_negated_dc_signal
			;

read_direct_dc_signal	:	read_identifier
					{AddToDontCare(DIRECT, $1); }
			;

read_negated_dc_signal	:	read_identifier
					 {AddToDontCare(NEGATED, $1); }
			;

list_of_dont_cares	: 	new_dont_care	   dont_care
			|	list_of_dont_cares new_dont_care dont_care
			;

new_dont_care		:	T_OR
					{NewDontCare(yylineno); }
			;

read_identifier		:	T_IDENTIFIER		{$$=$1;}
			;

%%
