/*********************************************************
 * project:     qmc -- minimize dnf using the Quine McCluskey algorithem
 * file:        select.c
 * version:     0.1
 * date:        july, 25. 1995
 * author:      volker bosch (volker.bosch@z.zgs.de)
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qmc.h"
#include "parserfctn.h"
#include "listfctn.h"
#include "select.h"
#include "minimize.h"


#ifdef	QMC_DEBUG
#include "debugfctn.h"
#endif


#define ERR_CODE_BASE	400


static void ClrNumberArray(int *Array, long Length) {
  long i;
  
  for(i = 0; i < Length; i++) {
    if(*Array > 0) *Array = 0;
    Array++;
  }
}

static void InitNumberArray(int *Array, long Length) {
  long i;
  
  for(i = 0; i < Length; i++) *Array++ = 0;
}


/*------------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static int AppearanceOfMinTerms(int *NumberArray, int NumberOfTerm) {
  int Dim, ArrayIdx, i;
  register char Mask;
  char *ArrayPtr;
  register MIN_TERM_LIST_ELEMENT *ListPtr;
  register int MaxNumberOfOneTerm;
  int Index;
  
  MaxNumberOfOneTerm = 0;
  ClrNumberArray(NumberArray, NumberOfTerm);
  for(Dim = 0; Dim < NumberOfSignals; Dim++) {
    ListPtr = MinTerms[Dim].StartOfListOfMinTerms;
    while(ListPtr!= NULL) {
      if((ListPtr->Flag != SELECT_FLAG) && (ListPtr->Flag != ESSENTIAL_FLAG)) {
        if((ArrayPtr = ListPtr->IdxArray) != NULL) {
          for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
            Mask = 1;
            for(i = 0; i < 8; i++) {
              if((*ArrayPtr & Mask) != 0) {
                Index = (ArrayIdx * 8) + i;
                if(Index <= NumberOfTerm) {
                  if(NumberArray[Index] >= 0) { 
                    NumberArray[Index]++;
                    if(NumberArray[Index] > MaxNumberOfOneTerm) 
                    		MaxNumberOfOneTerm = NumberArray[Index];
                  } /* if(NumberArray[Index] >= 0) */
                } /* if(Index <= NumberOfTerm) */
              } /* if((*ArrayPtr...*/
              Mask <<= 1;
            } /* for(i = 0 */
            ArrayPtr++;
          } /* for(ArrayIdx... */
        } /* (ListPtr->IdxArray != NULL) */
      } /* if((ListPtr->Flag != SELECT_FLAG) && ... */
      ListPtr = ListPtr->Successor;
    } /* while(ListPtr!= NULL) */
  } /* for(Dim...*/
  return(MaxNumberOfOneTerm);
}

/*--------------------------------------------
 * 
 *--------------------------------------------------*/
static int ValuateMinTerms(int Max, int *NumberArray, int NumberOfTerm) {
  int Dim, ArrayIdx, i;
  register char Mask;
  char *ArrayPtr;
  register MIN_TERM_LIST_ELEMENT *ListPtr;
  register int MaxValue;
  int Index;

  MaxValue = 0;
  for(Dim = 0; Dim < NumberOfSignals; Dim++) {
    ListPtr = MinTerms[Dim].StartOfListOfMinTerms;
    while(ListPtr != NULL) {
      if((ListPtr->Flag != SELECT_FLAG) && (ListPtr->Flag != ESSENTIAL_FLAG)) {
        ListPtr->Value = 0;
        if((ArrayPtr = ListPtr->IdxArray) != NULL) {
          for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
            Mask = 1;
            for(i = 0; i < 8; i++) {
              if((*ArrayPtr & Mask) != 0) {
                Index = (ArrayIdx * 8) + i;
                if(Index <= NumberOfTerm) {
                  if(NumberArray[Index] > 0) { 
                    ListPtr->Value += ((Max - NumberArray[Index]) + 1);
                    if(ListPtr->Value > MaxValue) MaxValue = ListPtr->Value;        
                  } /* if(NumberArray[Index] > 0) */
                } /* if(Index <= NumberOfTerm) */
              } /* if((*ArrayPtr...*/
              Mask <<= 1;
            } /* for(i = 0 */
            ArrayPtr++;
          } /* for(ArrayIdx... */
        } /* (ListPtr->IdxArray != NULL) */
      } /* if((ListPtr->Flag != SELECT_FLAG) && .. */
    ListPtr = ListPtr->Successor;
    } /* while(ListPtr!= NULL) */
  } /* for(Dim...*/
  return(MaxValue);
}

/*--------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void SelectMinTermOfValue(int Value, int *NumberArray, int NumberOfTerm) {
  int Dim, ArrayIdx, i;
  register char Mask;
  char *ArrayPtr;
  register MIN_TERM_LIST_ELEMENT *ListPtr;
  int Index;

  for(Dim = 0; Dim < NumberOfSignals; Dim++) {
    ListPtr = MinTerms[Dim].StartOfListOfMinTerms;
    while(ListPtr != NULL) {
      if((ListPtr->Flag != SELECT_FLAG) && (ListPtr->Flag != ESSENTIAL_FLAG)) {
        if(ListPtr->Value == Value) {
          ListPtr->Flag = SELECT_FLAG;
          ListPtr->Value = 0;
          if((ArrayPtr = ListPtr->IdxArray) != NULL) {
            for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
              Mask = 1;
              for(i = 0; i < 8; i++) {
                if((*ArrayPtr & Mask) != 0) {
                  Index = (ArrayIdx * 8) + i;
                  if(Index <= NumberOfTerm) {
                    NumberArray[Index] = -1;
                  } /* if(Index <= NumberOfTerm) */
                } /* if((*ArrayPtr...*/
                Mask <<= 1;
              } /* for(i = 0 */
              ArrayPtr++;
            } /* for(ArrayIdx... */
          } /* if((ArrayPtr = */
          return;
        } /* if(ListPtr->Value = Value) */
      } /* if((ListPtr->Flag != SELECT_FLAG) && .. */
    ListPtr = ListPtr->Successor;
    } /* while(ListPtr!= NULL) */
  } /* for(Dim...*/
}

/*--------------------------------------------------------
 *
 *-----------------------------------------------------------------------*/
static void SelectSingleTerms(int *NumberArray, int NumberOfTerm) {
  int Dim, ArrayIdx, i;
  register char Mask;
  char *ArrayPtr;
  register MIN_TERM_LIST_ELEMENT *ListPtr;
  int Index;
  int FoundSingleTerm;


  for(Dim = 0; Dim < NumberOfSignals; Dim++) {
    ListPtr = MinTerms[Dim].StartOfListOfMinTerms;
    while(ListPtr != NULL) {
      FoundSingleTerm = 0;
      if((ArrayPtr = ListPtr->IdxArray) != NULL) {
        for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
          Mask = 1;
          for(i = 0; i < 8; i++) {
            if((*ArrayPtr & Mask) != 0) {
              Index = (ArrayIdx * 8) + i;
              if(Index <= NumberOfTerm) {
                if(NumberArray[Index] == 1) {
                  ListPtr->Flag = ESSENTIAL_FLAG;
                  FoundSingleTerm = 1;
                } /* if(NumberArray[Index] == 1) */
              } /* if(Index <= NumberOfTerm) */
            } /* if((*ArrayPtr...*/
            Mask <<= 1;
          } /* for(i = 0 */
          ArrayPtr++;
        } /* for(ArrayIdx... */
        if(FoundSingleTerm == 1) {
          ArrayPtr = ListPtr->IdxArray;
          for(ArrayIdx = 0; ArrayIdx < SizeOfIdxArray; ArrayIdx++) {
            Mask = 1;
            for(i = 0; i < 8; i++) {
              if((*ArrayPtr & Mask) != 0) {
                Index = (ArrayIdx * 8) + i;
                if(Index <= NumberOfTerm) {
                  NumberArray[Index] = -1;
                } /* if(Index <= NumberOfTerm)  */
              } /* if((*ArrayPtr...*/
              Mask <<= 1;
            } /* for(i = 0 */
            ArrayPtr++;
          } /* for(ArrayIdx... */        
        } /* if(FoundSingleTerm == 1) */
      } /* if((ArrayPtr = ListPtr->IdxArray) != NULL) */
      ListPtr = ListPtr->Successor;
    } /* while(ListPtr!= NULL) */
  } /* for(Dim...*/
}

/*----------------------------------------------------
 *
 *---------------------------------------------------------------*/
static void MarkUnselectedTerms(void) {
  MIN_TERM_LIST_ELEMENT *List;
  int Dimension;  

  for(Dimension = 0; Dimension < NumberOfSignals; Dimension++) {
    List = MinTerms[Dimension].StartOfListOfMinTerms;
    while(List != NULL) {
      if((List->Flag != ESSENTIAL_FLAG) && (List->Flag != SELECT_FLAG)) {
         List->Flag = CLR_FLAG;
      } else {
         List->Flag = 0;
      }
      List = List->Successor;
    } /* while(List != NULL) */
  } /* for(Dimension ... */
}        

/*----------------------------------------------------
 *       
 *---------------------------------------------------------------*/
void SelectMinTerms(void) {
int *NumberOfMinTermIdx;
int MaxOfRemainingMinTerms;
int MaxValue;

  if((NumberOfMinTermIdx = (int *)calloc((unsigned int)SizeOfIdxArray * 8, sizeof(int))) == NULL) {
    fprintf(stderr,"calloc error in SelectMinTerms()\n");
    exit(ERR_CODE_BASE + 11);
  }
  AppearanceOfMinTerms(NumberOfMinTermIdx, MinTermIdx + 1);
  SelectSingleTerms(NumberOfMinTermIdx, MinTermIdx + 1);    
  while((MaxOfRemainingMinTerms = AppearanceOfMinTerms(NumberOfMinTermIdx, MinTermIdx + 1)) > 0) {
    MaxValue = ValuateMinTerms(MaxOfRemainingMinTerms, NumberOfMinTermIdx, MinTermIdx + 1);
    SelectMinTermOfValue(MaxValue, NumberOfMinTermIdx, MinTermIdx + 1);
  }
  /*-------------- remove unselected terms -------------------------*/
  MarkUnselectedTerms();
  ClearMarkedElements();  
  /*-------------- select 2nd time ---------------------------------*/
  InitNumberArray(NumberOfMinTermIdx, MinTermIdx + 1);
  AppearanceOfMinTerms(NumberOfMinTermIdx, MinTermIdx + 1);
  SelectSingleTerms(NumberOfMinTermIdx, MinTermIdx + 1);    
  while((MaxOfRemainingMinTerms = AppearanceOfMinTerms(NumberOfMinTermIdx, MinTermIdx + 1)) > 0) {
    MaxValue = ValuateMinTerms(MaxOfRemainingMinTerms, NumberOfMinTermIdx, MinTermIdx + 1);
    SelectMinTermOfValue(MaxValue, NumberOfMinTermIdx, MinTermIdx + 1);
  }  
  free(NumberOfMinTermIdx);
}
